﻿//--------------------------------------------------------------------------------------
// Copyright 2014-2015 Intel Corporation
// All Rights Reserved
//
// Permission is granted to use, copy, distribute and prepare derivative works of this
// software for any purpose and without fee, provided, that the above copyright notice
// and this statement appear in all copies.  Intel makes no representations about the
// suitability of this software for any purpose.  THIS SOFTWARE IS PROVIDED "AS IS."
// INTEL SPECIFICALLY DISCLAIMS ALL WARRANTIES, EXPRESS OR IMPLIED, AND ALL LIABILITY,
// INCLUDING CONSEQUENTIAL AND OTHER INDIRECT DAMAGES, FOR THE USE OF THIS SOFTWARE,
// INCLUDING LIABILITY FOR INFRINGEMENT OF ANY PROPRIETARY RIGHTS, AND INCLUDING THE
// WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.  Intel does not
// assume any responsibility for any errors which may appear in this software nor any
// responsibility to update it.
//--------------------------------------------------------------------------------------

using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Data;
using System.Windows.Documents;
using System.Windows.Input;
using System.Windows.Media;
using System.Windows.Media.Imaging;
using System.Windows.Navigation;
using System.Windows.Shapes;
using System.Threading;

namespace Sketch
{
    /// <summary>
    /// Interaction logic for MainWindow.xaml
    /// </summary>
    public partial class MainWindow : Window
    {
        private enum Gesture
        {
            Undefined,
            FingerSpread,
            Pinch,
            Wave
        };

        private enum PenColor
        {
            Undefined,
            Red,
            Yellow,
            Green,
            Blue,
            Violet,
            Orange,
            Brown,
            Black
        };

        private Gesture gesture;
        private Thread processingThread;
        private PXCMSenseManager senseManager;
        private Int32 nhands;
        private Int32 handId;
        private float handTipX;
        private float handTipY;
        private float handTipZ;
        private string detectionAlert;
        private string calibrationAlert;
        private string bordersAlert;
        private bool detectionStatusOk;
        private bool calibrationStatusOk;
        private bool borderStatusOk;
        private PXCMHandModule hand;
        private PXCMHandData handData;
        private PXCMHandData.GestureData gestureData;
        private PXCMHandData.IHand ihand;
        private PXCMHandData.JointData[][] nodes;
        private PXCMHandConfiguration handConfig;
        private const int MaxPointSize = 30;
        private const int DrawingCanvasHeight = 600;
        private const int DrawingCanvasWidth = 600;
        private const int arraySize = 30;
        private float[] xValues;
        private float[] yValues;
        private float[] zValues;
        private int xyzValueIndex;
        Point currentPoint;
        Brush ColorBrush;

        public MainWindow()
        {
            InitializeComponent();
            nodes = new PXCMHandData.JointData[][] { new PXCMHandData.JointData[0x20], new PXCMHandData.JointData[0x20] };
            currentPoint = new Point();
            ChangePenColor(PenColor.Red);
            gesture = Gesture.Undefined;
            xValues = new float[arraySize];
            yValues = new float[arraySize];
            zValues = new float[arraySize];
            xyzValueIndex = 0;
            detectionAlert = string.Empty;
            calibrationAlert = string.Empty;
            bordersAlert = string.Empty;
            detectionStatusOk = false;
            calibrationStatusOk = false;
            borderStatusOk = false;
            processingThread = new Thread(new ThreadStart(ProcessingThread));
            senseManager = PXCMSenseManager.CreateInstance();
            senseManager.EnableHand();
            senseManager.Init();
            ConfigureHandModule();
            processingThread.Start();
        }

        private void ConfigureHandModule()
        {
            hand = senseManager.QueryHand();
            handConfig = hand.CreateActiveConfiguration();
            handConfig.EnableGesture("spreadfingers");
            handConfig.EnableGesture("two_fingers_pinch_open");
            handConfig.EnableGesture("wave");
            handConfig.EnableAllAlerts();
            handConfig.ApplyChanges();
        }

        private void Window_Loaded(object sender, RoutedEventArgs e)
        {
            drawingCanvas.Height = DrawingCanvasHeight;
            drawingCanvas.Width = DrawingCanvasWidth;
            ellDrawingPoint.Height = 10;
            ellDrawingPoint.Width = 10;
            Canvas.SetTop(ellDrawingPoint, DrawingCanvasHeight / 2 - ellDrawingPoint.Height / 2);
            Canvas.SetRight(ellDrawingPoint, DrawingCanvasWidth / 2 - ellDrawingPoint.Width / 2);
        }

        private void Window_Closing(object sender, System.ComponentModel.CancelEventArgs e)
        {
            processingThread.Abort();
            if (handData != null) handData.Dispose();
            if (handConfig != null) handConfig.Dispose();
            senseManager.Dispose();
        }

        private void ProcessingThread()
        {
            // Start AcquireFrame/ReleaseFrame loop
            while (senseManager.AcquireFrame(true) >= pxcmStatus.PXCM_STATUS_NO_ERROR)
            {
                hand = senseManager.QueryHand();

                if (hand != null)
                {

                    // Retrieve the most recent processed data
                    handData = hand.CreateOutput();
                    handData.Update();

                    // Get number of tracked hands
                    nhands = handData.QueryNumberOfHands();

                    if (nhands > 0)
                    {
                        // Retrieve hand identifier
                        handData.QueryHandId(PXCMHandData.AccessOrderType.ACCESS_ORDER_BY_TIME, 0, out handId);

                        // Retrieve hand data
                        handData.QueryHandDataById(handId, out ihand);

                        // Retrieve all hand joint data
                        for (int i = 0; i < nhands; i++)
                        {
                            for (int j = 0; j < 0x20; j++)
                            {
                                PXCMHandData.JointData jointData;
                                ihand.QueryTrackedJoint((PXCMHandData.JointType)j, out jointData);
                                nodes[i][j] = jointData;
                            }
                        }

                        // Get world coordinates for tip of middle finger on the first hand in camera range
                        handTipX = nodes[0][Convert.ToInt32(PXCMHandData.JointType.JOINT_MIDDLE_TIP)].positionWorld.x;
                        handTipY = nodes[0][Convert.ToInt32(PXCMHandData.JointType.JOINT_MIDDLE_TIP)].positionWorld.y;
                        handTipZ = nodes[0][Convert.ToInt32(PXCMHandData.JointType.JOINT_MIDDLE_TIP)].positionWorld.z;

                        // Retrieve gesture data
                        if (handData.IsGestureFired("spreadfingers", out gestureData)) { gesture = Gesture.FingerSpread; }
                        else if (handData.IsGestureFired("two_fingers_pinch_open", out gestureData)) { gesture = Gesture.Pinch; }
                        else if (handData.IsGestureFired("wave", out gestureData)) { gesture = Gesture.Wave; }
                    }
                    else
                    {
                        gesture = Gesture.Undefined;
                    }

                    // Get alert status
                    for (int i = 0; i < handData.QueryFiredAlertsNumber(); i++)
                    {
                        PXCMHandData.AlertData alertData;
                        if (handData.QueryFiredAlertData(i, out alertData) != pxcmStatus.PXCM_STATUS_NO_ERROR) { continue; }

                        //Displaying last alert
                        switch (alertData.label)
                        {
                            case PXCMHandData.AlertType.ALERT_HAND_DETECTED:
                                detectionAlert = "Hand Detected";
                                detectionStatusOk = true;
                                break;
                            case PXCMHandData.AlertType.ALERT_HAND_NOT_DETECTED:
                                detectionAlert = "Hand Not Detected";
                                detectionStatusOk = false;
                                break;
                            case PXCMHandData.AlertType.ALERT_HAND_CALIBRATED:
                                calibrationAlert = "Hand Calibrated";
                                calibrationStatusOk = true;
                                break;
                            case PXCMHandData.AlertType.ALERT_HAND_NOT_CALIBRATED:
                                calibrationAlert = "Hand Not Calibrated";
                                calibrationStatusOk = false;
                                break;
                            case PXCMHandData.AlertType.ALERT_HAND_INSIDE_BORDERS:
                                bordersAlert = "Hand Inside Borders";
                                borderStatusOk = true;
                                break;
                            case PXCMHandData.AlertType.ALERT_HAND_OUT_OF_BORDERS:
                                bordersAlert = "Hand Out Of Borders";
                                borderStatusOk = false;
                                break;
                        }
                    }

                    UpdateUI();
                    if (handData != null) handData.Dispose();
                }
                senseManager.ReleaseFrame();
            }
        }

        private void UpdateUI()
        {
            this.Dispatcher.Invoke(System.Windows.Threading.DispatcherPriority.Normal, new Action(delegate()
            {
                // Show x-y hand coordinates
                lblCoordinates.Content = string.Format("X: {0},   Y: {1}", handTipX, handTipY);
                 
                // Do moving average filtering
                if (xyzValueIndex < arraySize)
                {
                    xValues[xyzValueIndex] = handTipX;
                    yValues[xyzValueIndex] = handTipY;
                    zValues[xyzValueIndex] = handTipZ;
                    xyzValueIndex++;
                }
                else
                {
                    for (int i = 0; i < arraySize - 1; i++)
                    {
                        xValues[i] = xValues[i + 1];
                        yValues[i] = yValues[i + 1];
                        zValues[i] = zValues[i + 1];
                    }

                    xValues[arraySize - 1] = handTipX;
                    yValues[arraySize - 1] = handTipY;
                    zValues[arraySize - 1] = handTipZ;
                }
                
                // Update Z axis ellipse
                float scaledZ = zValues.Average() * -100 + 50;

                if (scaledZ > MaxPointSize)
                {
                    ellDrawingPoint.Height = MaxPointSize;
                    ellDrawingPoint.Width = MaxPointSize;
                }
                else if (scaledZ < 0)
                {
                    ellDrawingPoint.Height = 0;
                    ellDrawingPoint.Width = 0;
                }
                else
                {
                    ellDrawingPoint.Height = scaledZ;
                    ellDrawingPoint.Width = scaledZ;
                }
                
                // Update drawing canvas
                Canvas.SetRight(ellDrawingPoint, (xValues.Average() * 3000 + 300) - ellDrawingPoint.Width / 2);
                Canvas.SetTop(ellDrawingPoint, (yValues.Average() * -3000 + 300) - ellDrawingPoint.Height / 2);

                // Draw line
                if ((gesture == Gesture.Pinch) && (handTipX >= -0.12))
                {
                    ellDrawingPoint.Stroke = ColorBrush;
                    ellDrawingPoint.Fill = ColorBrush;
                    Line line = new Line();
                    line.Stroke = ColorBrush;
                    line.StrokeThickness = scaledZ;
                    line.StrokeDashCap = PenLineCap.Round;
                    line.StrokeStartLineCap = PenLineCap.Round;
                    line.StrokeEndLineCap = PenLineCap.Round; 
                    line.X1 = currentPoint.X;
                    line.Y1 = currentPoint.Y;
                    line.X2 = xValues.Average() * -3000 + 300;
                    line.Y2 = yValues.Average() * -3000 + 300;

                    currentPoint.X = xValues.Average() * -3000 + 300;
                    currentPoint.Y = yValues.Average() * -3000 + 300;
                    drawingCanvas.Children.Add(line);
                }
                else
                {
                    ellDrawingPoint.Stroke = ColorBrush;
                    ellDrawingPoint.Fill = Brushes.Transparent;
                    currentPoint.X = xValues.Average() * -3000 + 300;
                    currentPoint.Y = yValues.Average() * -3000 + 300;
                }

                // Erase canvas on hand wave
                if (gesture == Gesture.Wave)
                {
                    drawingCanvas.Children.Clear();
                    drawingCanvas.Children.Add(ellDrawingPoint);
                }

                // Handle gesture-based color menu selections
                if (handTipX < -0.12)
                {
                    if (gesture == Gesture.FingerSpread)
                    {
                        if ((handTipY <= 0.1) && (handTipY > 0.075))
                        {
                            ChangePenColor(PenColor.Red);
                        }
                        else if ((handTipY <= 0.075) && (handTipY > 0.05))
                        {
                            ChangePenColor(PenColor.Yellow);
                        }
                        else if ((handTipY <= 0.05) && (handTipY > 0.025))
                        {
                            ChangePenColor(PenColor.Green);
                        }
                        else if ((handTipY <= 0.025) && (handTipY > 0.0))
                        {
                            ChangePenColor(PenColor.Blue);
                        }
                        else if ((handTipY <= 0.0) && (handTipY > -0.025))
                        {
                            ChangePenColor(PenColor.Violet);
                        }
                        else if ((handTipY <= -0.025) && (handTipY > -0.05))
                        {
                            ChangePenColor(PenColor.Orange);
                        }
                        else if ((handTipY <= -0.05) && (handTipY > -0.075))
                        {
                            ChangePenColor(PenColor.Brown);
                        }
                        else if ((handTipY <= -0.075) && (handTipY > -0.1))
                        {
                            ChangePenColor(PenColor.Black);
                        }
                    }
                }

                // Update gesture info
                switch (gesture)
                {
                    case Gesture.Undefined:
                        lblDraw.Foreground = Brushes.White;
                        lblErase.Foreground = Brushes.White;
                        lblHover.Foreground = Brushes.White;
                        break;
                    case Gesture.FingerSpread:
                        lblDraw.Foreground = Brushes.White;
                        lblErase.Foreground = Brushes.White;
                        lblHover.Foreground = Brushes.LightGreen;
                        break;
                    case Gesture.Pinch:
                        lblDraw.Foreground = Brushes.LightGreen;
                        lblErase.Foreground = Brushes.White;
                        lblHover.Foreground = Brushes.White;
                        break;
                    case Gesture.Wave:
                        lblDraw.Foreground = Brushes.White;
                        lblErase.Foreground = Brushes.LightGreen;
                        lblHover.Foreground = Brushes.White;
                        break;
                }
                
                // Update alert info
                lblDetectionAlert.Foreground = (detectionStatusOk) ? System.Windows.Media.Brushes.LightGreen : System.Windows.Media.Brushes.Red;
                lblDetectionAlert.Content = detectionAlert;

                lblCalibrationAlert.Foreground = (calibrationStatusOk) ? System.Windows.Media.Brushes.LightGreen : System.Windows.Media.Brushes.Red;
                lblCalibrationAlert.Content = calibrationAlert;

                lblBordersAlert.Foreground = (borderStatusOk) ? System.Windows.Media.Brushes.LightGreen : System.Windows.Media.Brushes.Red;
                lblBordersAlert.Content = bordersAlert;
            }));
        }

        private void btnRed_Click(object sender, RoutedEventArgs e)
        {
            ChangePenColor(PenColor.Red);
        }

        private void btnYellow_Click(object sender, RoutedEventArgs e)
        {
            ChangePenColor(PenColor.Yellow);
        }

        private void btnGreen_Click(object sender, RoutedEventArgs e)
        {
            ChangePenColor(PenColor.Green);
        }

        private void btnBlue_Click(object sender, RoutedEventArgs e)
        {
            ChangePenColor(PenColor.Blue);
        }

        private void btnViolet_Click(object sender, RoutedEventArgs e)
        {
            ChangePenColor(PenColor.Violet);
        }

        private void btnOrange_Click(object sender, RoutedEventArgs e)
        {
            ChangePenColor(PenColor.Orange);
        }

        private void btnBrown_Click(object sender, RoutedEventArgs e)
        {
            ChangePenColor(PenColor.Brown);
        }

        private void btnBlack_Click(object sender, RoutedEventArgs e)
        {
            ChangePenColor(PenColor.Black);
        }

        private void ChangePenColor(PenColor color)
        {
            switch (color)
            {
                case PenColor.Red:
                    ColorBrush = Brushes.Red;
                    borderRed.BorderBrush = Brushes.White;
                    borderYellow.BorderBrush = Brushes.Gray;
                    borderGreen.BorderBrush = Brushes.Gray;
                    borderBlue.BorderBrush = Brushes.Gray;
                    borderViolet.BorderBrush = Brushes.Gray;
                    borderOrange.BorderBrush = Brushes.Gray;
                    borderBrown.BorderBrush = Brushes.Gray;
                    borderBlack.BorderBrush = Brushes.Gray;
                    break;
                case PenColor.Yellow:
                    ColorBrush = Brushes.Yellow;
                    borderRed.BorderBrush = Brushes.Gray;
                    borderYellow.BorderBrush = Brushes.White;
                    borderGreen.BorderBrush = Brushes.Gray;
                    borderBlue.BorderBrush = Brushes.Gray;
                    borderViolet.BorderBrush = Brushes.Gray;
                    borderOrange.BorderBrush = Brushes.Gray;
                    borderBrown.BorderBrush = Brushes.Gray;
                    borderBlack.BorderBrush = Brushes.Gray;
                    break;
                case PenColor.Green:
                    ColorBrush = Brushes.Green;
                    borderRed.BorderBrush = Brushes.Gray;
                    borderYellow.BorderBrush = Brushes.Gray;
                    borderGreen.BorderBrush = Brushes.White;
                    borderBlue.BorderBrush = Brushes.Gray;
                    borderViolet.BorderBrush = Brushes.Gray;
                    borderOrange.BorderBrush = Brushes.Gray;
                    borderBrown.BorderBrush = Brushes.Gray;
                    borderBlack.BorderBrush = Brushes.Gray;
                    break;
                case PenColor.Blue:
                    ColorBrush = Brushes.Blue;
                    borderRed.BorderBrush = Brushes.Gray;
                    borderYellow.BorderBrush = Brushes.Gray;
                    borderGreen.BorderBrush = Brushes.Gray;
                    borderBlue.BorderBrush = Brushes.White;
                    borderViolet.BorderBrush = Brushes.Gray;
                    borderOrange.BorderBrush = Brushes.Gray;
                    borderBrown.BorderBrush = Brushes.Gray;
                    borderBlack.BorderBrush = Brushes.Gray;
                    break;
                case PenColor.Violet:
                    ColorBrush = Brushes.Violet;
                    borderRed.BorderBrush = Brushes.Gray;
                    borderYellow.BorderBrush = Brushes.Gray;
                    borderGreen.BorderBrush = Brushes.Gray;
                    borderBlue.BorderBrush = Brushes.Gray;
                    borderViolet.BorderBrush = Brushes.White;
                    borderOrange.BorderBrush = Brushes.Gray;
                    borderBrown.BorderBrush = Brushes.Gray;
                    borderBlack.BorderBrush = Brushes.Gray;
                    break;
                case PenColor.Orange:
                    ColorBrush = Brushes.Orange;
                    borderRed.BorderBrush = Brushes.Gray;
                    borderYellow.BorderBrush = Brushes.Gray;
                    borderGreen.BorderBrush = Brushes.Gray;
                    borderBlue.BorderBrush = Brushes.Gray;
                    borderViolet.BorderBrush = Brushes.Gray;
                    borderOrange.BorderBrush = Brushes.White;
                    borderBrown.BorderBrush = Brushes.Gray;
                    borderBlack.BorderBrush = Brushes.Gray;
                    break;
                case PenColor.Brown:
                    ColorBrush = Brushes.Brown;
                    borderRed.BorderBrush = Brushes.Gray;
                    borderYellow.BorderBrush = Brushes.Gray;
                    borderGreen.BorderBrush = Brushes.Gray;
                    borderBlue.BorderBrush = Brushes.Gray;
                    borderViolet.BorderBrush = Brushes.Gray;
                    borderOrange.BorderBrush = Brushes.Gray;
                    borderBrown.BorderBrush = Brushes.White;
                    borderBlack.BorderBrush = Brushes.Gray;
                    break;
                case PenColor.Black:
                    ColorBrush = Brushes.Black;
                    borderRed.BorderBrush = Brushes.Gray;
                    borderYellow.BorderBrush = Brushes.Gray;
                    borderGreen.BorderBrush = Brushes.Gray;
                    borderBlue.BorderBrush = Brushes.Gray;
                    borderViolet.BorderBrush = Brushes.Gray;
                    borderOrange.BorderBrush = Brushes.Gray;
                    borderBrown.BorderBrush = Brushes.Gray;
                    borderBlack.BorderBrush = Brushes.White;
                    break;
                default:
                    break;
            }
        }
    }
}
